#!/usr/bin/env python

# Copyright by Risto A. Paju (teknohog at iki.fi) 2014-2015

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

#######################################################################

resfiles = {"finished": "client_finished.txt",
            "solution": "client_solutions.txt",
            }

sentfile = "client_sent.txt"
statefile = "client_state.txt"

worker_bins = {"julia": "worker.jl",
               "python": "worker.py",
      }

host = "kasj.tunk.org"

import subprocess

import random, urllib2, os, sys
import argparse

from threading import Thread, enumerate as thread_enum, Event
from Queue import Queue
from time import sleep, time

# Constants
from dnals_common import quantum, infimum, wulimits, http_port, random_overstep

# Functions
from dnals_common import wu_generate, wu_parse, text_decode, text_encode, write_file, file_readlines, solution_verify

# Waits/timeouts in seconds
delay = {"progress": 2,
         "save": 300,
     }

def random_work():
    global random_overstep, wulimits

    # Besides network outages, this function is needed when the server
    # runs out of work. So go beyond the limits, but not too far.

    wu = map(lambda x: random.randrange(x + 1), wulimits)

    if random_overstep:
        # Overstep in only one dimension
        i = random.randrange(2)
        wu[i] += wulimits[i]
        
    return wu_generate(wu)

def cleanup_finished(fin_str):
    # If we continue after an interruption, the finish code will not
    # contain the original range, so fix it here. Whether this happens
    # can only be deduced from the numbers (statefile already has
    # "finished" by then), so we need a little trust here.

    # Note, the state variable is now xf rather than n.

    global quantum, infimum

    # text_decode already has try/except protections, but need to
    # check for list length
    finlist = text_decode(fin_str)

    if len(finlist) != 5:
        # Let it fall through for subsequent logging
        return fin_str
    
    # This is not a full validity check, it will be done elsewhere for
    # all outcomes anyway.
    if finlist[0] == infimum[0] or finlist[0] % quantum[0] == 0:
        # No modification needed.
        return fin_str
    else:
        # The int() is not necessary in Python, but it makes the
        # down-rounding logic clear
        finlist[0] = max(infimum[0], int(finlist[0]/quantum[0]) * quantum[0])

        fin_str = text_encode(finlist)
        return fin_str

def network_getwork(getwork):
    # Continuing from unfinished work is handled elsewhere. This
    # function now only sends results from files and gets new work.
    
    global resfiles, sentfile, net, baseurl

    # Initialize these every time (duh)
    url = baseurl
    work = ""
    
    # The keys also serve as markers for files to clean up, and whether we
    # have anything to send at all
    sendlist = {}

    for k in resfiles:
        if os.path.exists(resfiles[k]):
            sendlist[k] = file_readlines(resfiles[k])

            if k == "finished":
                sendlist[k] = map(cleanup_finished, sendlist[k])
            
            url += query_generate(k, sendlist[k])

    if getwork == False:
        if len(sendlist) == 0:
            # Don't bother the server for nothing
            return work
        else:
            url += query_generate("nowork", ["true"])
            
    try:
        print("Fetching " + url)

        r = net.open(url)
        output = r.read()

        print("Got " + output)
    
        if len(wu_parse(output)) == 2:
            # For now, we only fetch new work after finishing the old one,
            # so overwrite.
            work = output

        # At this point we have submitted results, so can delete these and
        # log them away
        for k in sendlist:
            os.remove(resfiles[k])
            write_file(sentfile, sendlist[k])
        
    except:
        print("Fetch error")

    if len(work) == 0 and getwork:
        work = random_work()
        print("No work available, generated random " + work)

    return work

# For a single variable type, repeated many times
def query_generate(name, values):
    output = ""
    for v in values:
        output += name + "=" + v + "&"

    return output

def init_state(statefile, args):
    # Startup: check statefile for interrupted work, get new work as
    # necessary.

    nworkers = args.threads
    getwork = not args.nowork
    
    states = []
    mtimes = []
    
    for ext in ["", ".bak"]:
        filename = statefile + ext
        try:
            states.append(file_readlines(filename))
            mtimes.append(os.path.getmtime(filename))
        except:
            pass

    # 2015-01-30 Choose the latest file. Ideally, we should also check
    # for validity of state, but this should be fine for now.
    
    if len(states) > 0:
        #numworks = map(len, states)
        #i = numworks.index(max(numworks))
        i = mtimes.index(max(mtimes))
        works = states[i]
    else:
        works = []

    # Choose valid works by reordering, don't just copy the old list
    # and fill in the blanks (see the point below about a full-length
    # list).
        
    new_works = []

    for w in works:
        if len(wu_parse(w, True)) == 2 and len(new_works) < nworkers:
            print("Continuing old work with " + w)
            new_works.append(w)

    # We may just want to finish off old works, so keep this optional,
    # no need for a full-length list of (potentially empty) works
    if getwork:
        while len(new_works) < nworkers:
            new_works.append(network_getwork(True))
    else:
        # We may still have some unsent results from previous runs
        network_getwork(False)
            
    return new_works
        
#######################

class Worker(Thread):
    def __init__(self, lang, work):
        global worker_bins
        
        Thread.__init__(self)

        self.daemon = True

        self.name = "worker"
        
        self.process_list = [lang, worker_bins[lang], work]

        # Disable buffering in the output of Python workers, otherwise
        # we cannot monitor/manage them. Julia seems to do this right
        # by default, but there may still be issues left, depending on
        # the OS etc.
        if lang == "python":
            self.process_list.insert(1, "-u")
        
        # Written by this thread only, I guess it is safe to read by others
        self.state = work

        self.work_numbers = text_decode(self.state)

        # Used by progressbar
        self.xmin = self.work_numbers[0]
        self.xmax = self.work_numbers[1]
        
    def progress_str(self):
        # text_decode has its own try/except protections
        self.work_numbers = text_decode(self.state)
        if len(self.work_numbers) > 0:
            x = self.work_numbers[0]
            return str(round(float(x - self.xmin) / (self.xmax - self.xmin) * 100, 2)) + " %"
        else:
            return "stopped"
        
    def run(self):
        global result_queue
        
        self.process = subprocess.Popen(self.process_list, stdout = subprocess.PIPE)
        
        while True:
            out = self.process.stdout.readline()
            
            if out == '' or self.process.poll() != None:
                # This probably signals a worker error...
                break
            elif out.startswith("finished"):
                result_queue.put(out)
                break
            elif out.startswith("solution"):
                result_queue.put(out)
            else:
                # This should only contain a raw work/state, thus ok
                # to text_decode
                self.state = out

        # Save a blank state, rather than keeping the old one
        self.state = ""
                
class Progressbar(Thread):
    def __init__(self):
        Thread.__init__(self)

        self.daemon = True
        
    def run(self):
        global worker_threads, delay
        
        while True:
            # Don't assume a fixed number of workers
            progs = [w.progress_str() for w in worker_threads]
            
            print("DNaLS thread progress: " + " ".join(progs))
            
            sleep(delay["progress"])

class State_saver(Thread):
    def __init__(self):
        Thread.__init__(self)

        self.daemon = True
        
    def run(self):
        global worker_threads, delay, statefile, ss_event
        
        while True:
            for ext in ["", ".bak"]:
                # Save upon startup, so the new getworks from a fresh
                # start are safe.
            
                # Don't assume a fixed number of workers
                states = [w.state.strip() for w in worker_threads]
                s = "\n".join(states)
            
                write_file(statefile + ext, s, "w")

                print("Saved state\n" + s)

                ss_event.wait(delay["save"])
                ss_event.clear()
            
class Main_loop(Thread):
    def __init__(self):
        Thread.__init__(self)

        self.daemon = True
        
    def run(self):
        global args, result_queue, resfiles, worker_threads, ss_event
        
        # We don't need file locking if all file operations are handled here
        # in one sequential loop.
        while True:
            # Get results from workers into the appropriate files

            # A single queue for both kinds of results beats separate ones in
            # many ways :)
            res = result_queue.get()
    
            getwork = False

            for k in resfiles:
                if res.startswith(k):
                    resdata = res.split()[-1]
                    write_file(resfiles[k], resdata)
        
                    if k == "finished" and not args.nowork:
                        getwork = True
                    elif k == "solution":
                        print("Found solution " + resdata)
                
            result_queue.task_done()

            # General note: we keep the old way of moving data via
            # files, in case of network outages etc. We can
            # automatically pick up from the old files after a
            # restart.
            
            # Send the results in files and optionally get new work (a single
            # WU for now).
            work = network_getwork(getwork)

            # Launch/restart the worker thread with the new work. __init__
            # seems to work, and it avoids both recreating new threads, and
            # writing custom startup/initialization functions within.
            if len(work) > 0:
                for w in worker_threads:
                    if not w.isAlive():
                        w.__init__(lang, work)
                        w.start()
                        break
            # We could do this more intelligently by using each
            # worker's individual ident, passed along with the
            # result. However, I don't see a direct way to call a
            # thread with the given ident without going through the
            # whole list, so this is as good as any.

            # Overwrite the finished states / save the freshly got work units
            ss_event.set()
            
            # If there is no work left after the last network_getwork, quit.
            tnames = [t.name for t in thread_enum()]
            if "worker" not in tnames:
                print("Out of work.")
                return

parser = argparse.ArgumentParser()
parser.add_argument("-n", "--nowork", action="store_true", help="Finish current work, report it and exit")
parser.add_argument("-o", "--host", type=str, default = "http://" + host + ":" + str(http_port) + "/", help="Alternative host URL, such as http://kasj.tunk.org/~teknohog/math/dnals/getwork.php")
parser.add_argument("-p", "--pyworker", action="store_true", help="Use the Python GMPY2 worker instead of Julia")
parser.add_argument("-t", "--threads", type=int, help="Number of worker threads", default=1)
parser.add_argument("-u", "--user", type=str, required=True, help="Your username")
args = parser.parse_args()

if args.pyworker:
    lang = "python"
else:
    lang = "julia"
            
result_queue = Queue()

ss_event = Event()

net = urllib2.build_opener()

baseurl = args.host + "?user=" + args.user + "&lang=" + lang + "&"

# Initialize workers with possible unfinished states

works = init_state(statefile, args)

if len(works) == 0:
    print("Nothing to do.")
    sys.exit()

worker_threads = []
for w in works:
    worker_threads.append(Worker(lang, w))
        
for w in worker_threads:
    w.start()

# Get the daemons up
    
progbar = Progressbar()
progbar.start()

statesaver = State_saver()
statesaver.start()

# Rinse and repeat with new work

ml = Main_loop()
ml.start()

try:
    while True:
        # Threads are generally hard to interrupt. So they are left
        # running as daemons, and we do something simple here that can
        # be easily terminated to bring down the entire script.
        if not ml.isAlive():
            break
        else:
            # Or whatever delay you fancy
            sleep(delay["save"])
        
except KeyboardInterrupt:
    print("Terminated")
                    
