#!/usr/bin/env julia

# Copyright by Risto A. Paju (teknohog at iki.fi) 2014-2015

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

#######################################################################

# 10 decimals of e
checksum_mod = 7182818284

function root_find_diff(x, n, qkrs)
    s = 0
    checksum = 0
    
    for qkr in qkrs
        # Theorem NL, part (iv)
        if qkr[1] > 3 || (n - 5) % 6 == 0
        # Test known solutions
        #if 0 == 0

            fp = fermat_pretest(x, n, qkr)
            if fp == 0
                # The same s can be used for all q, so compute only once
                if s == 0
                    bx = BigInt(x)
                    s = BigInt(round((bx^n - 1) / (bx - 1)))
                end
            
                by = BigInt(round(s^(1/qkr[1])))
                
                byq = by^qkr[1]
                
                diff = s - byq

                # The parallelized client will now collect solutions
                # whenever they appear, so no need to wait any further in
                # the overall loop
                if diff == 0
                    solution = text_encode([x, by, n, qkr[1]])
                    
                    print("solution ", solution, "\n")
                    
                    # This should be fairly unique to avoid parallel
                    # clashes, but the entire solution might be too long
                    # for a filename
                    sol_id = (x + by + n + qkr[1]) % 4294967296
                    solfile = string("solution-", sol_id, ".txt")
                    write_file(solfile, solution)
                else
                    checksum += diff
                end
            else
                checksum += fp
            end
        end
    end
        
    return checksum
end

function write_file(filename, l, mode="a")
    # Append mode is best for results, thus default
    content = string(l, "\n")
    File = open(filename, mode)
    write(File, content)
    close(File)
end

# Dead simple, but I want to keep these abstract for flexibility. The 
# dashed format now matches the distributed setup.
function text_encode(l)
    return join(l, "-")
end

function text_decode(t)
    return map(int, split(t, "-"))
end

function nfilter(n)
    # As explained in PredaYann1.ps et al.
    
    fd = factor(n)

    # The dictionary design is broken beyond all hope if I have to
    # filter these myself to see what the actual KEYS and VALUES are,
    # instead of some random memory locations. The seemingly random
    # bug annoyed me for a good while before this amazing revelation
    # that fd.slots shows where they are. If this weren't a potential
    # school project, I'd be swearing my posterior off.

    exponents =  map(*, fd.slots, fd.vals)
    bases = map(*, fd.slots, fd.keys)

    if minimum(filter(x -> x > 0, bases)) < 29
        # The smallest prime divisor should be at least 29. Also, I
        # heard you like filters so I put a filter in my filter so we
        # can filter while we filter. Also, Julia's min() is for
        # distinct numbers and minimum() for arrays.
        return false
    elseif sum(exponents) > 3
        # The remaining primes are now at least 29. There should be
        # at most 3 of these, counting multiplicities.
        return false
    else
        return true
    end
end

function qkrlist(x)
    fd = factor(x)

    # See nfilter() why we need this
    bases = map(*, fd.slots, fd.keys)

    pmlist = filter(x -> x > 0, bases) - 1

    # Potential divisors of pmlist; q > 2
    qprimes = primes(int(maximum(pmlist)/2) + 1)[2:end]

    # Test known solutions with q = 2 allowed
    #qprimes = primes(int(maximum(pmlist)/2) + 1)
    
    qkrs = Tuple[]

    # We only need a list of the q's. No need to find all (p, q) pairs
    # as earlier, hence the break.
    
    for q in qprimes
        for pm in pmlist
            if pm % q == 0
                # Take this q, and find r and k such that
                # * r is prime
                # * r = k*q + 1
                # * r is not a factor of x

                # Starting with a fixed list of primes r will easily
                # end up in errors. This may not be optimal, but at
                # least we find the smallest possible r.

                k = 2
                r = k*q + 1

                # 2015-01-27 improvement: find multiple suitable k, r
                # pairs for a stronger pre-bignum test.

                krs = Tuple[]

                # Anything more than about 20 has little effect, as we
                # hardly ever get into bignums
                while size(krs)[1] < 20
                    if isprime(r) && !(r in bases)
                        push!(krs, (k, r))
                    end

                    k += 1
                    r = k*q + 1
                end
                
                push!(qkrs, (q, krs))
                break
            end
        end
    end

    return qkrs
end

function fermat_pretest(x, n, qkr)
    # As suggested by R. Gerbicz, to avoid going into bignums in most cases

    # A true solution will pass all test rounds; any single failure
    # means no solution. Use 0 as true, so we can collect checksums
    # here.

    # If r divides x^n - 1, the test won't work, so these cases must
    # be counted as passes.
    
    for kr in qkr[2]
        xnr1 = powermod(x, n, kr[2]) - 1
        if xnr1 != 0 && powermod(xnr1, kr[1], kr[2]) != powermod(x-1, kr[1], kr[2])
            return kr[2]
        end
    end

    return 0
end
    
xmin, xmax, nmin, nmax, checksum = text_decode(ARGS[1])

# Report the state at reasonable intervals
report_int = max(2, int(round((xmax - xmin) / 1000)))

# Avoid overlap and consider the ranges of xf and n in Python style --
# i.e. xmax and nmax are never quite reached.
nlist = filter(nfilter, nmin:(nmax - 1))

# Test known results below the limit of 29
#nlist = nmin:(nmax - 1)

for x = xmin:(xmax - 1)
    if x % report_int == 0
        # Fix the potentially negative modular checksum
        checksum += checksum_mod
        workstate = text_encode([x, xmax, nmin, nmax, checksum])
        print(workstate, "\n")
    end

    qkrs = qkrlist(x)

    if size(qkrs)[1] > 0
        for n in nlist
            # Solutions are now handled within root_find_diff
            cs = root_find_diff(x, n, qkrs)
            checksum = (checksum + cs) % checksum_mod
        end
    end
end

# In Julia, the modular checksum can still be negative, so fix it
# before the final result
checksum = (checksum + checksum_mod) % checksum_mod

finished = text_encode([xmin, xmax, nmin, nmax, checksum])

print("finished ", finished, "\n")
