#!/usr/bin/env python

# Copyright by Risto A. Paju (teknohog at iki.fi) 2014-2015

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

#######################################################################

# 10 decimals of e
checksum_mod = 7182818284

import sys, itertools, bisect, gmpy2

# Functions
from dnals_common import text_decode, text_encode, write_file

def primes(n):
    "List primes from 2 up to n, using the Sieve of Eratosthenes"

    # Avoid marks and other extra lists, for less memory and a
    # possible speedup: Mark multiples by setting them to zero, remove
    # them as we go on, but not all at once, as we can use properly
    # spaced indexing. But in fact this is much slower, because the
    # pop() causes us to shuffle around a huge list. Using the same
    # list as markers is a sensible idea, though.

    # list() is needed for python3, otherwise this is a range object
    integers = list(range(int(n)+1))

    primes = []

    # This looks a little dangerous, though it works, since the list
    # is read before running the loop.
    for i in integers[2:]:
        if integers[i] > 0:
            primes.append(i)
            for j in range(2*i, len(integers), i):
                integers[j] = 0

    return primes

def root_find_diff(x, n, qkrs):
    s = 0
    checksum = 0
    
    for qkr in qkrs:
        # Theorem NL, part (iv)
        if qkr[0] > 3 or (n - 5) % 6 == 0:
        # Test known solutions
        #if 0 == 0:

            fp = fermat_pretest(x, n, qkr)
            if fp == 0:
                # The same s can be used for all q, so compute only once
                if s == 0:
                    bx = gmpy2.mpz(x)
                    s = (pow(bx, n) - 1) / (bx - 1)
            
                by = gmpy2.mpz(s**(1./qkr[0]))
    
                byq = by**qkr[0]
            
                diff = s - byq

                # The parallelized client will now collect solutions
                # whenever they appear, so no need to wait any further in
                # the overall loop
                if diff == 0:
                    solution = text_encode([x, by, n, qkr[0]])
                    
                    print("solution " + solution)
                    
                    # This should be fairly unique to avoid parallel
                    # clashes, but the entire solution might be too long
                    # for a filename
                    sol_id = (x + by + n + qkr[0]) % 4294967296
                    solfile = "solution-" + str(sol_id) + ".txt"
                    write_file(solfile, solution)
                else:
                    checksum += diff
            else:
                checksum += fp
                    
    return checksum

def factor(n):
    # Julia has a relatively handy factor(n) so do something similar
    # with a dictionary.

    if n < 2:
        return {n: 1}
    
    flist = {}

    for p in primes(n**0.5 + 1):
        while n % p == 0:
            n /= p

            if p in flist:
                flist[p] += 1
            else:
                flist[p] = 1

    # Anything that remains is 1 or a single prime
    if n > 1:
        flist[n] = 1
                
    return flist

def nfilter(n):
    global nprimes
    
    # As explained in PredaYann1.ps et al.
    
    # The smallest prime divisor should be at least 29
    
    for sp in nprimes[:9]:
        if n % sp == 0:
            return False

    # There should be at most 3 prime divisors, counting
    # multiplicities.  29 is the 10th prime, or index 9 in Pythonese.

    # 29^3 - take the easy way out :)
    if n <= 24389:
        return True
    
    factors = 0

    # We can optimize the full factorization a little, since we only
    # need to find a limited number of factors
        
    for lp in nprimes[9:]:
        while n % lp == 0:
            n /= lp
            factors += 1

            if factors > 3:
                return False

    # Anything that remains is 1 or a single prime
    if n > 1 and factors > 2:
        return False
    else:
        return True

def qkrlist(x):
    # We need a full factorization here to find the big factor > x**0.5
    fd = factor(x)

    pmlist = map(lambda x: x - 1, fd.keys())
    
    # Potential divisors of pmlist; q > 2
    qprimes = primes(max(pmlist)/2 + 1)[1:]

    # Test known results with q = 2 allowed
    #qprimes = primes(max(pmlist)/2 + 1)
    
    qkrs = []

    # We only need a list of the q's. No need to find all (p, q) pairs
    # as earlier, hence the break.
    
    for q in qprimes:
        for pm in pmlist:
            if pm % q == 0:
                # Take this q, and find r and k such that
                # * r is prime
                # * r = k*q + 1
                # * r is not a factor of x

                # Starting with a fixed list of primes r will easily
                # end up in errors. This may not be optimal, but at
                # least we find the smallest possible r.

                k = 2
                r = k*q + 1

                # 2015-01-27 improvement: find multiple suitable k, r
                # pairs for a stronger pre-bignum test.

                krs = []

                # Anything more than about 20 has little effect, as we
                # hardly ever get into bignums
                while len(krs) < 20:
                    if is_prime(r) and r not in fd.keys():
                        krs.append((k, r))

                    k += 1
                    r = k*q + 1

                qkrs.append((q, krs))
                break

    return qkrs

def is_prime(n):
    # Start with something known/fast
    global testprimes, max_testprime

    if n in testprimes:
        return True

    for p in testprimes:
        if n % p == 0:
            return False

    # If the list runs out, use this idea adapted from
    # http://stackoverflow.com/questions/15285534/isprime-function-for-python-language
    r = int(n**0.5)
    f = max_testprime
    while f <= r:
        if n%f == 0: return False
        if n%(f+2) == 0: return False
        f +=6

    return True    

def fermat_pretest(x, n, qkr):
    # As suggested by R. Gerbicz, to avoid going into bignums in most cases

    # A true solution will pass all test rounds; any single failure
    # means no solution. Use 0 as true, so we can collect checksums
    # here.

    # If r divides x^n - 1, the test won't work, so these cases must
    # be counted as passes.
    
    for kr in qkr[1]:
        xnr1 = pow(x, n, kr[1]) - 1
        if xnr1 != 0 and pow(xnr1, kr[0], kr[1]) != pow(x-1, kr[0], kr[1]):
            return kr[1]

    return 0

# Dashed format for the distributed setup
xmin, xmax, nmin, nmax, checksum = text_decode(sys.argv[1])

from time import time
starttime = time()

# For is_prime(). Eats up a huge amount of memory, but there's not
# much point in optimizing the Python client too far, when we have
# Julia too.
# A set is faster when matching if n is there, but then it
# is not ordered, and we need to keep the max somehow
testprimes = primes(5*xmax)
max_testprime = testprimes[-1]
testprimes = set(testprimes)

# Avoid overlap and consider the ranges of xf and n in Python style --
# i.e. xfmax and nmax are never quite reached.

# For nfilter
nprimes = primes(nmax**0.5 + 1)

nlist = filter(nfilter, range(nmin, nmax+1))

# Test known results below the limit of 29
#nlist = range(nmin, nmax+1)

# Report the state at reasonable intervals
report_int = max(2, int(round((xmax - xmin) / 1000)))

for x in range(xmin, xmax):
    if x % report_int == 0:
        workstate = text_encode([x, xmax, nmin, nmax, checksum])
        print(workstate)
        
    qkrs = qkrlist(x)

    if len(qkrs) > 0:
        for n in nlist:
            # Solutions are now handled within root_find_diff
            cs = root_find_diff(x, n, qkrs)
            checksum = (checksum + cs) % checksum_mod

finished = text_encode([xmin, xmax, nmin, nmax, checksum])

print("finished " + finished)
